--[[
**********************************************************************
MagicDKP Client - a client for the Magic DKP System.
**********************************************************************
This file is part of MagicDKP Client, a World of Warcraft Addon

MagicDKP Client is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

MagicDKP Client is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP Client.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]

MDKP_Client = LibStub("AceAddon-3.0"):NewAddon("MagicDKP_Client", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local mod = MDKP_Client
MagicDKP_ClientDB = {}

local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP_Client", false)
local GUI = LibStub("AceGUI-3.0")
local MagicComm = LibStub("MagicComm-1.0")
local media = LibStub("LibSharedMedia-3.0")
local currentEvent, currentSender, currentRaid
local fmt = string.format
local networkData = {}
local pendingStandby = {}
local playerName, playerClass
local wipe = table.wipe
local GetItemInfo = GetItemInfo
local select = select
local tostring = tostring
local tonumber = tonumber
local db = MagicDKP_ClientDB
local time = time
local pairs = pairs
local match = string.match
local IsEquippableItem = IsEquippableItem
local defaults = {
   sound = "Cartoon FX", 
   bidSound = "Loot Chime",
   filterUnusable = true
}

-- Used to hide bids for items the players class can't use
local classRoles = {
   DEATHKNIGHT = { spell = false, melee = true,  tank = true,  useInt = false, useStr = true  },
   DRUID       = { spell = true,  melee = true,  tank = true,  useInt = true,  useStr = true  },
   MAGE        = { spell = true,  melee = false, tank = false, useInt = true,  useStr = false },
   PALADIN     = { spell = true,  melee = true,  tank = true,  useInt = true,  useStr = true  },
   PRIEST      = { spell = true,  melee = false, tank = false, useInt = true,  useStr = false },
   ROGUE       = { spell = false, melee = true,  tank = false, useInt = false, useStr = false },
   SHAMAN      = { spell = true,  melee = true,  tank = false, useInt = true,  useStr = true  },
   WARLOCK     = { spell = true,  melee = false, tank = false, useInt = true,  useStr = false },
   WARRIOR     = { spell = false, melee = true,  tank = true,  useInt = false, useStr = true  },
   HUNTER      = { spell = false, melee = true,  tank = false, useInt = false, useStr = false },
}

mod.MAJOR_VERSION = "MagicDKP_Client-1.0"
mod.MINOR_VERSION = tonumber('63') or tonumber(("$Revision: 62 $"):match("%d+"))

function mod:OnInitialize()
   media:Register("sound", "Bats", [[Sound\Doodad\BatsFlyAway.wav]]) 
   media:Register("sound", "Cartoon FX", [[Sound\Doodad\Goblin_Lottery_Open03.wav]]) 
   media:Register("sound", "Cheer", [[Sound\Event Sounds\OgreEventCheerUnique.wav]]) 
   media:Register("sound", "Clockwork", [[Sound\Doodad\G_GasTrapOpen.wav]]) 
   media:Register("sound", "Drop", [[Sound\Interface\DropOnGround.wav]]) 
   media:Register("sound", "Error", [[Sound\Interface\Error.wav]]) 
   media:Register("sound", "Explosion", [[Sound\Doodad\Hellfire_Raid_FX_Explosion05.wav]]) 
   media:Register("sound", "Fel Nova", [[Sound\Spells\SeepingGaseous_Fel_Nova.wav]]) 
   media:Register("sound", "Fel Portal", [[Sound\Spells\Sunwell_Fel_PortalStand.wav]]) 
   media:Register("sound", "Firework", [[Sound\Doodad\G_FireworkBoomGeneral2.wav]]) 
   media:Register("sound", "Fog Horn", [[Sound\Doodad\ZeppelinHorn.wav]]) 
   media:Register("sound", "Friend Login", [[Sound\Interface\FriendJoin.wav]]) 
   media:Register("sound", "Gong", [[Sound\Doodad\G_GongTroll01.wav]]) 
   media:Register("sound", "Humm", [[Sound\Spells\SimonGame_Visual_GameStart.wav]]) 
   media:Register("sound", "Loot Chime", [[Sound\Interface\igLootCreature.wav]]) 
   media:Register("sound", "Magic Click", [[Sound\Interface\MagicClick.wav]]) 
   media:Register("sound", "Ping", [[Sound\Interface\MapPing.wav]]) 
   media:Register("sound", "Rubber Ducky", [[Sound\Doodad\Goblin_Lottery_Open01.wav]]) 
   media:Register("sound", "Shing!", [[Sound\Doodad\PortcullisActive_Closed.wav]]) 
   media:Register("sound", "Short Circuit", [[Sound\Spells\SimonGame_Visual_BadPress.wav]]) 
   media:Register("sound", "Simon Chime", [[Sound\Doodad\SimonGame_LargeBlueTree.wav]]) 
   media:Register("sound", "Socket Clunk", [[Sound\Interface\JewelcraftingFinalize.wav]]) 
   media:Register("sound", "War Drums", [[Sound\Event Sounds\Event_wardrum_ogre.wav]]) 
   media:Register("sound", "Wham!", [[Sound\Doodad\PVP_Lordaeron_Door_Open.wav]]) 
   media:Register("sound", "Whisper Ping", [[Sound\Interface\iTellMessage.wav]]) 
   media:Register("sound", "Whisper", [[Sound\Interface\igTextPopupPing02.wav]]) 
   media:Register("sound", "Wisp", [[Sound\Event Sounds\Wisp\WispPissed1.wav]]) 
   media:Register("sound", "You Will Die!", [[Sound\Creature\CThun\CThunYouWillDIe.wav]]) 

   playerName = UnitName("player")
   playerClass = select(2, UnitClass("player"))

   for id, val in pairs(defaults) do
      if db[id] == nil then db[id] = val end
   end

  StaticPopupDialogs["MDKPStandbyCheck"] = {
     text = "",
     button1 = L["I am here!"],
     whileDead = 1,
     hideOnEscape = 0,
     timeout = 300,
     startDelay = 0,
     delayText = "Hello",
     OnAccept = function() mod:SendStandbyAck() mod:ScheduleTimer("NextCheck", 2) end,
     OnShow = function()
		 mod.popupTimer = mod:ScheduleRepeatingTimer("UpdatePopupTimer", 1, { time(), getglobal(this:GetName().."Text") })
	      end
  }
  mod:InitializeTooltip()

   
   mod:CreateBidDialog()
   mod.CreateBidDialog = nil
   mod.InitializeTooltip = nil
end

function mod:OnEnable()
   MagicComm:RegisterListener(mod, "MD")
   mod:RegisterOptions()
end

function mod:OnDisable()
   MagicComm:UnregisterListener(mod, "MD")
end

local function SetNetworkData(cmd, data, misc1, misc2, misc3, misc4)
   networkData.cmd = cmd
   networkData.data = data
   networkData.misc1 = misc1
   networkData.misc2 = misc2
   networkData.misc3 = misc3
   networkData.misc4 = misc4
end

function mod:SendUrgentMessage(channel, recipient)
   MagicComm:SendUrgentMessage(networkData, "MD", channel, recipient)
end

local standbyText = L["%s has initiated a standby check for %s. Standby check expiring in %ds."]

function mod:UpdatePopupTimer(data)
   local timeleft = 300 - (time()-data[1])
   if timeleft > 0 then
      data[2]:SetText(standbyText:format(currentSender, currentEvent, timeleft))
   else
      mod:CancelTimer(mod.standbyTimer, true)
      StaticPopup_Hide("MDKPStandbyCheck")
      currentSender = nil
      currentEvent = nil
      currentRaid = nil
      mod:ScheduleTimer("NextCheck", 2)
   end
end


local bidFormat = L["Taking bids for %s"]
local bidFormatTimer = L["Taking bids for %s\nBidding closes in %ds."]

local function BidDialog_Cleanup()
   mod.bidDialog:Hide() 
   mod.selectedPriority = nil
   mod.item = nil
   mod:CancelTimer(mod.bidTimerScheduler, true)
   mod.bidTimeExpire = nil
   mod.bidTimerScheduler = nil
   mod.bidmaster = nil
end

local function BidDialog_OnAccept(self)
   local dkp = tonumber(mod.bidDialog:GetValue())
   if not mod.selectedPriority or mod.selectedPriority > #mod.priorities or mod.selectedPriority < 1 then
      mod:Print(L["You need to select the bid priority!"])
   else
      SetNetworkData("DKPRESPONSE", playerName, dkp, mod.selectedPriority, mod.item)
      mod:SendUrgentMessage("WHISPER", mod.bidmaster)
      BidDialog_Cleanup()
   end
end			 

local function BidDialog_Pass()
   SetNetworkData("DKPRESPONSE", playerName, nil, nil, mod.item)
   mod:SendUrgentMessage("WHISPER", mod.bidmaster)
   BidDialog_Cleanup()
end

local function BidDialog_AutoPass()
   SetNetworkData("DKPRESPONSE", playerName, -1, nil, mod.item)
   mod:SendUrgentMessage("WHISPER", mod.bidmaster)
   BidDialog_Cleanup()
end

function mod:CreateBidDialog()
   local dialog =  GUI:Create("MDKP_BidDialog")
   dialog:SetTitle(L["New Item Bid"])
   dialog:SetCallback("OnBidPass", BidDialog_Pass)
   dialog:SetCallback("OnAccept", BidDialog_OnAccept)
   dialog:SetCallback("OnPrioritySelected", function(self, event, value) mod.selectedPriority = value end)
   dialog:Hide()
   mod.bidDialog = dialog
end

function mod:UpdateBidTimer()
   if not mod.bidTimeExpire then return end
   local timeleft = mod.bidTimeExpire - time()
   if timeleft > 0 then
      if not mod.bidTimerScheduler then
	 mod.bidTimerScheduler = mod:ScheduleRepeatingTimer("UpdateBidTimer", 1)
      end
      mod.bidDialog:SetText(bidFormatTimer:format(mod.item, timeleft))
   else
      BidDialog_Pass()
   end
end

function mod:OnDKPBalanceResponse(data)
   mod.dkpBalance = data
end

-- Time to bid on an item
function mod:OnDKPBid(item, priorities, bidmaster, expirationTime, dkpdata)
   local dialog = mod.bidDialog
   mod.item = item
   mod.priorities = priorities
   mod.bidmaster = bidmaster
   mod.selectedPriority = 1
   local info = L["DKP Balance: %s"]:format(L["Unknown"])
   if mod.dkpBalance then
      local current
      if dkpdata then 
	 current = mod.dkpBalance[dkpdata.multipool.."_current"]
      else
	 current = mod.dkpBalance.DKP
      end
      if current then
	 info = L["DKP Balance: %s"]:format(tostring(current))
      end
   end

   if dkpdata and dkpdata.num then
      info = info ..", "..fmt(L["Value: %d-%d ~%d in %d drops"], dkpdata.min, dkpdata.max, dkpdata.avg, dkpdata.num)
   end
   
   PlaySoundFile(media:Fetch("sound", db.bidSound))
   if db.filterUnusable and not mod:IsItemUsable(item) then
      mod:Print(L["Auto-passed bid for %s since you can't use it."]:format(item))
      BidDialog_AutoPass()
      return
   end

   if expirationTime then
      mod.bidTimeExpire = time() + expirationTime
   end

   dialog:SetInfoText(info or "")
   dialog:SetValue(0)
   dialog:SetPriorityList(priorities)
   dialog:SetItemLink(item)
   dialog:Show()
   dialog.select.frame:Show()
   mod.bidDialog:SetText(bidFormat:format(mod.item))
   mod:UpdateBidTimer()
end

function mod:OnDKPCancel(item, bidmaster)
   if mod.item == item and mod.bidmaster == bidmaster then
      --      mod:Print("The current bid for", item,"has been ended by", bidmaster)
      BidDialog_Pass()
   end
end

function mod:OnStandbyCheck(event, sender, raidid)
   if currentEvent then
      pendingStandby[#pendingStandby+1] = {
	 sender = sender,
	 event = event,
	 raid = raidid,
      }
   else
      mod:CancelAllTimers()
      mod:ShowStandbyCheckDialog(sender, event, raidid)
   end
end


function mod:ShowStandbyCheckDialog(sender, event, raidid)
   currentSender = sender
   currentEvent = event
   currentRaid = raidid
   StaticPopupDialogs["MDKPStandbyCheck"].text = standbyText:format(currentSender, currentEvent, 300)
   StaticPopup_Show("MDKPStandbyCheck")
   PlaySoundFile(media:Fetch("sound", db.sound))
end

function mod:SendStandbyAck()
   mod:CancelAllTimers()
   StaticPopup_Hide("MDKPStandbyCheck")
   if currentSender and currentEvent then
      SetNetworkData("STANDBYRESPONSE", playerName, currentEvent, currentRaid)
      mod:SendUrgentMessage("WHISPER", currentSender)
      currentSender = nil
      currentEvent = nil
      currentRaid = nil
   end
end

function mod:NextCheck()
   if currentEvent then return end
   local last = #pendingStandby
   if last > 0 then
      mod:ShowStandbyCheckDialog(pendingStandby[last].sender, 
				 pendingStandby[last].event, 
				 pendingStandby[last].raid)
      pendingStandby[last] = nil
   end
end

-- This will wait for the server to refresh the information about
-- the item after setting the tooltip

function mod:RefreshItemData(item)
   local name = GetItemInfo(item)
   if not name then
      self:ScheduleTimer("RefreshItemData", 0.5, item)
      return
   end
   if not mod:IsItemUsable(item, true) then
      BidDialog_AutoPass()
   end
   return
end

do
   local colorFormat = "%02X%02X%02X"
   local errorColor = "FE1F1F"
   local text = {}
   local TT = CreateFrame("GameTooltip")
   local casterPatterns = {}
   local meleePatterns = {}
   local tankPatterns = {}
   local statPatterns = { }

   local meleeStats = {
      ITEM_MOD_AGILITY = true, 
      ITEM_MOD_ARMOR_PENETRATION_RATING = true, 
      ITEM_MOD_ATTACK_POWER = true, 
      ITEM_MOD_EXPERTISE_RATING = true, 
      ITEM_MOD_RANGED_ATTACK_POWER = true, 
      ITEM_MOD_STRENGTH = true, 
      ITEM_MOD_AGILITY_SHORT = true, 
      ITEM_MOD_ARMOR_PENETRATION_RATING_SHORT = true, 
      ITEM_MOD_ATTACK_POWER_SHORT = true, 
      ITEM_MOD_EXPERTISE_RATING_SHORT = true, 
      ITEM_MOD_RANGED_ATTACK_POWER_SHORT = true, 
      ITEM_MOD_STRENGTH_SHORT = true, 
   }
   
   local tankStats = {
      ITEM_MOD_BLOCK_RATING = true, 
      ITEM_MOD_DEFENSE_SKILL_RATING = true, 
      ITEM_MOD_DODGE_RATING = true, 
      ITEM_MOD_PARRY_RATING = true, 
      ITEM_MOD_BLOCK_RATING_SHORT = true, 
      ITEM_MOD_DEFENSE_SKILL_RATING_SHORT = true, 
      ITEM_MOD_DODGE_RATING_SHORT = true, 
      ITEM_MOD_PARRY_RATING_SHORT = true, 
   }
   
   local casterStats = {
      ITEM_MOD_SPELL_DAMAGE_DONE = true, 
      ITEM_MOD_SPELL_HEALING_DONE = true, 
      ITEM_MOD_SPELL_POWER = true, 
      ITEM_MOD_SPIRIT = true, 
      ITEM_MOD_SPELL_DAMAGE_DONE_SHORT = true, 
      ITEM_MOD_SPELL_HEALING_DONE_SHORT = true, 
      ITEM_MOD_SPELL_POWER_SHORT = true, 
      ITEM_MOD_SPIRIT_SHORT = true, 
   }

   function mod:InitializeTooltip()
      TT:SetOwner(UIParent, "ANCHOR_NONE")
      for i=1,60,2 do
	 text[i], text[i+1] = TT:CreateFontString(), TT:CreateFontString()
	 text[i]:SetFontObject(GameFontNormal)
	 text[i+1]:SetFontObject(GameFontNormal)
	 TT:AddFontStrings(text[i], text[i+1])
      end
   end


   local function ClearTooltip()
      TT:ClearLines()
      for l = 2,60,2 do text[l]:SetText() end
   end

   -- This method checks for the red "not usable" color on the first 5 lines.
   -- Should match all Class: XXX and items not equippable by the player.
   local function LowIsItemUsable(item)
      local r, g, b
      for l = 1, 2*math.min(TT:NumLines(), 10) do
	 if text[l]:GetText() then
	    r,g,b = text[l]:GetTextColor()
	    if colorFormat:format(r*255, g*255, b*255) == errorColor then
	       return false
	    end
	 end
      end
      return true
   end

   function mod:IsItemUsable(item, cached)
      local hasSpell, hasTank, hasMelee
      ClearTooltip()
      TT:SetHyperlink(item)
      if not cached and not (GetItemInfo(item)) then
	 mod:RefreshItemData(item)
	 return true -- defer autopass
      end
      if not LowIsItemUsable(item) then return false end
      if not IsEquippableItem(item) then return true end -- not limiting non-equippable item by class

      -- loop through all item stats and detect caster, melee and tank stats
      local itemStats = GetItemStats(item)
      for stat in pairs(itemStats) do
	 	--mod:Print("Checking stat ", stat)
	 if casterStats[stat] then
	    hasSpell = true
	    --mod:Print("Spell stat: ", stat);
	 elseif meleeStats[stat] then
	    hasMelee = true
	    --mod:Print("Melee stat: ", stat);
	 elseif tankStats[stat] then 
	    hasTank = true
	    --mod:Print("Tank stat: ", stat);
	 end
      end
      
      local pc = classRoles[playerClass]
      if itemStats.ITEM_MOD_INTELLECT and not pc.useInt then
	 -- in some cases, "non-int" users use int. Specifically if the item has
	 -- melee stats. For the sake of simplicity in code, we count hunters as
	 -- not using int. Rogues never use int but warriors and DK's can sometimes
	 -- use mail items with int.
	 if hasMelee and pc.melee and playerClass ~= "HUNTER" then
	    itemStats.ITEM_MOD_INTELLECT = nil
	 end
      end
      return not (   hasSpell and not pc.spell
		  or hasMelee and not pc.melee
		  or hasTank  and not pc.tank
		  or itemStats.ITEM_MOD_STRENGTH and not pc.useStr
		  or itemStats.ITEM_MOD_INTELLECT and not pc.useInt
	       )
   end
end

function mod:SetProfileParam(var, value, sub)
   db[var[#var]] = value
end

function mod:GetProfileParam(var, sub)
   return db[var[#var]]
end

local config = {
   type = "group",
   name = L["MagicDKP Client"],
   handler = mod,
   set = "SetProfileParam",
   get = "GetProfileParam",
   args = {
      sound = {
         type = "select",
         dialogControl = "LSM30_Sound",
         name = L["Standby Check Sound"],
         desc = L["Sound to play when a ready check is called"],
         values = AceGUIWidgetLSMlists.sound,
	 order = 1, 
      },
      bidSound= {
         type = "select",
         dialogControl = "LSM30_Sound",
         name = L["DKP Bid Sound"],
         desc = L["Sound to play when a new DKP bid is received"],
         values = AceGUIWidgetLSMlists.sound,
	 order = 1, 
      },
      filterUnusable = {
	 type = "toggle",
	 name = L["Filter bids for unusable items"],
	 desc = L["Don't show a DKP bid popup for items your class can't use."], 
	 width = "full"
      }
   }
}

function mod:RegisterOptions()
   LibStub("AceConfig-3.0"):RegisterOptionsTable(L["MagicDKP Client"], config)
   LibStub("AceConfigDialog-3.0"):AddToBlizOptions(L["MagicDKP Client"], L["MagicDKP Client"])
end

-- /dump MDKP_Client:OnStandbyCheck("Event", "Helaani")
